

lfp_sccs([], _).
lfp_sccs([SCC | SCCs], Trace) :-
	format("~w ", [SCC]), ttyflush,
	lfp_scc(SCC, 1, Trace),
	lfp_sccs(SCCs, Trace).

lfp_scc(Index, 1, Trace) :-
	integer(Index),
	!,
	lfp_scc_pass([Index], 1, Trace).
lfp_scc(Indices, Pass, Trace) :-
	lfp_scc_pass(Indices, Pass, Trace),
	retract(redo),
	!,
	NextPass is Pass + 1,
	lfp_scc(Indices, NextPass, Trace).
lfp_scc(_, _, _).
	
lfp_scc_pass([], _, _).
lfp_scc_pass([Index | Indices], Pass, Trace) :-
	!,
	(lfp_clause(Index, Pass, Trace); true),
	lfp_scc_pass(Indices, Pass, Trace).

lfp_clause(Index, Pass, Trace) :-
        abstract:head_to_linear_clause(Index, Atom, Atoms), 
        lfp_atoms(Atoms, Scalars),
	Atom =.. [_ | Args],
	project(Scalars, Args, Project),
	widen_scalars(Project, 4, WidenProject),
        cond_assert(Atom, WidedProject, Trace),
        fail.

lfp_atoms([], []).
lfp_atoms([Atom | Atoms], [Atom | Scalars]) :- 
	functor(Atom, scalar, 3),
	!,
	lfp_atoms(Atoms, Scalars).
lfp_atoms([Atom | Atoms], Scalars) :- 
	linear_atom(Atom, Scalars, RestScalars, _),
	lfp_atoms(Atoms, RestScalars).

widen_scalars([], _, []).
widen_scalars([scalar(Sum, RelOp, Value) | Scalars], Threshold, WidenScalars) :-
	abs(Value) =< Threshold,
	not_widen_sum(Sum, Threshold),
	!,
	WidenScalars = [scalar(Sum, RelOp, Value) | RestWidenScalars],
	widen_scalars(Scalars, Threshold, RestWidenScalars).
widen_scalars([_ | Scalars], Threshold, WidenScalars) :-
	widen_scalars(Scalars, Threshold, WidenScalars).

not_widen_sum([], _).
not_widen_sum([_-Coeff | Sum], Threshold) :-
	abs(Coeff) =< Threshold,
	not_widen_sum(Sum, Threshold).

stepper(off, _, _, _).
stepper(step, Message, Atom, Scalars) :-
	stepper_aux(log, Message, Atom, Scalars),

redo_assert :-
	redo, !.
redo_assert :-
	asserta(redo).

cond_assert(Atom, Scalars, Trace) :-
	linear_atom(Atom, FixScalars, [], HistoryScalars), 
	!,
	cond_assert_aux(Atom, Scalars, FixScalars, HistoryScalars, Pass, Trace).
cond_assert(Atom, Scalars, Trace) :-
	stepper(Trace, new, Atom, Scalars),
	redo_assert,
	append(Scalars, Scalars_In, Scalars_Out),
	asserta(linear_atom(Atom, Scalars_Out, Scalars_In, Scalars)). 

cond_assert_aux(Atom, Scalars, FixScalars, _, Pass, Trace) :-
	stepper_aux(Trace, fix_entails(Pass), Atom, FixScalars),
	stepper(Trace, new_entails(Pass), Atom, Scalars),
	entails(Scalars, FixScalars),
%	format("~nentailment holds", []), ttyflush,
	!.
cond_assert_aux(Atom, Scalars, FixScalars, HistoryScalars, Pass, Trace) :-
	Atom =.. [Name | Args],
	stepper_aux(Trace, lub, Atom, Scalars),
	stepper_aux(Trace, fix_lub, Atom, FixScalars),
	convex_hull(Args, Scalars, Args, FixScalars, LubArgs, LubScalars),
	LubAtom =.. [Name | LubArgs],
	LubAtom = Atom,
	stepper_aux(Trace, result_lub, Atom, LubScalars),
	redo_assert,
	retract(linear_atom(Atom, _, _, _)),
	append(LubScalars, Scalars_In, Scalars_Out),
	append(LubScalars, HistoryScalars, NewHistory),
	sort(NewHistory, SortedHistory),
	asserta(linear_atom(Atom, Scalars_Out, Scalars_In, SortedHistory)). 

