%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%:- eval('example.pl').
%:- eval('ex1.pl').

%fact(split0([],_,[],[]), [], 9, _).
%fact(split0([pair(A,B)|_],pair(C,_),[pair(A,B)|_],_), [A<C], 6, _).
%fact(split0([pair(A,B)|_],pair(C,B),_,_), [A=:=C], 7, _).
%fact(split0([pair(A,B)|_],pair(C,_),_,[pair(A,B)|_]), [A>C], 8, _).

generate_singleton_positions(0, Poss) :-
	!,
	Poss = [].
generate_singleton_positions(Pos, [[Pos]|Poss]) :-
	Pos1 is Pos - 1,
	generate_singleton_positions(Pos1, Poss).

union_positions(Positions, Union) :-
	map_map_union_positions(Positions, Positions, [], Union).

map_map_union_positions([], _, Union, Union).
map_map_union_positions([Pos1 |Poss1], Poss2, Union1, Union3) :-
	map_union_positions(Poss2, Pos1, Union1, Union2),
	map_map_union_positions(Poss1, Poss2, Union2, Union3).

map_union_positions([], _, Union, Union).
map_union_positions([Pos2 | Poss2], Pos1, Union1, Union2) :-
	ord_union(Pos1, Pos2, PosUnion),
	length(Pos1, LengthIn),
	length(PosUnion, LengthOut),
	LengthOut =:= LengthIn + 1,
	\+ memberchk(PosUnion, Union1),
	!,
	map_union_positions(Poss2, Pos1, [PosUnion|Union1], Union2).
map_union_positions([_ | Poss2], Pos1, Union1, Union2) :-
	map_union_positions(Poss2, Pos1, Union1, Union2).

map_subset([Pair | _], Triple) :-
	ord_subset(Pair, Triple),
	!.
map_subset([_ | Pairs], Triple) :-
	map_subset(Pairs, Triple).

remove_supersets([], _, []).
remove_supersets([Triple | Triples], Pairs, Residual) :-
	map_subset(Pairs, Triple),
	!,
	remove_supersets(Triples, Pairs, Residual).
remove_supersets([Triple | Triples], Pairs, [Triple | Residual]) :-
	remove_supersets(Triples, Pairs, Residual).

sufficient_conditions(Name, Arity, [[]]) :-
	definitely_disjoint([], Name, Arity),
	!.
sufficient_conditions(Name, Arity, Conds) :-
	generate_singleton_positions(Arity, Positions),
	map_definitely_disjoint(Positions, Name, Arity, Conds1, Residual),
	union_positions(Residual, Pairs),
	map_definitely_disjoint(Pairs, Name, Arity, Conds2, ResidualPairs),
	union_positions(ResidualPairs, Triples),
	remove_supersets(Triples, Conds2, FilteredTriples),
	map_definitely_disjoint(FilteredTriples, Name, Arity, Conds3, _),
	append(Conds1, Conds2, Conds12),
	append(Conds12, Conds3, Conds).

map_definitely_disjoint([], _Name, _Arity, [], []).
map_definitely_disjoint([Pos|Poss], Name, Arity, [Pos|Conds], Residual) :-
	definitely_disjoint(Pos, Name, Arity),
	!,
	map_definitely_disjoint(Poss, Name, Arity, Conds, Residual).
map_definitely_disjoint([Pos|Poss], Name, Arity, Conds, [Pos|Residual]) :-
	map_definitely_disjoint(Poss, Name, Arity, Conds, Residual).

definitely_disjoint(Positions, Name, Arity) :-
	functor(Atom, Name, Arity),
	findall(Index, fact(Atom, _, Index, _), Indices),
	map_map_definitely_disjoint_pair(Indices, Indices, Positions).

map_map_definitely_disjoint_pair([], _, _).
map_map_definitely_disjoint_pair([Index1 | Indices1], Indices2, Positions) :-
	map_definitely_disjoint_pair(Indices2, Index1, Positions),
	map_map_definitely_disjoint_pair(Indices1, Indices2, Positions).

map_definitely_disjoint_pair([], _, _).
map_definitely_disjoint_pair([Index2 | Indices2], Index1, Positions) :-
	Index1 == Index2,
	!,
	map_definitely_disjoint_pair(Indices2, Index1, Positions).
map_definitely_disjoint_pair([Index2 | Indices2], Index1, Positions) :-
	definitely_disjoint_pair(Positions, Index2, Index1),
	map_definitely_disjoint_pair(Indices2, Index1, Positions).

definitely_disjoint_pair(Positions, Index1, Index2) :-
	\+ possibly_not_disjoint_pair(Positions, Index1, Index2).

unify_args([], _, _).
unify_args([Position | Positions], Atom1, Atom2) :-
	arg(Position, Atom1, Arg),
	arg(Position, Atom2, Arg),
	unify_args(Positions, Atom1, Atom2).

possibly_not_disjoint_pair(Positions, Index1, Index2) :-
	fact(Atom1, Cons1, Index1, _),
	fact(Atom2, Cons2, Index2, _),
	unify_args(Positions, Atom1, Atom2),
	append(Cons1, Cons2, Cons),
	possibly_satisfiable_cons(Cons),
	possibly_satisfiable_cons_pair(Cons).


